package ru.kfu.itis;

import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.time.LocalTime;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import static java.lang.System.exit;

public class Server implements Runnable {
    private static int port = 0;
    private static String citiesFile = null;
    private Socket _socket = null;
    private BufferedReader _in;
    private PrintWriter _out;
    private StringBuffer _lastTown = null;
    private static Set _towns = new HashSet<String>();
    private Set availableTowns = new HashSet<String>();
    private  Set playedTowns = new HashSet<String>();
    private int _score = 0;
    private boolean terminated = false;
    private Character _lastChar = Character.MIN_VALUE;;

    public Server(Socket socket){
        _socket = socket;
    }

    public static void main(String[] args){

        if (args.length <2) {
            printInfo("\n Usage:\n\n java Main.class port 'citiesFile'");
            exit(0);
        }
        port = Integer.parseInt(args[0]);
        citiesFile = args[1];

        if(loadSet(_towns, citiesFile)){
            printInfo("Loaded "+_towns.size()+" cities from file "+citiesFile+"\n\n");
        }

        try (ServerSocket s = new ServerSocket(port);) {
            printInfo("Server started at "+s.getInetAddress()+":"+s.getLocalPort()+"\n\n");
            while(true) {
                Socket incoming = s.accept();
                printInfo("Connection accepted from "+incoming.getInetAddress()+" at "+ LocalTime.now()+"\n\n");
                new Thread(new Server(incoming)).start();
            }
        } catch (IOException ioex){
            printInfo("Server failed listening port: "+port+"\n\n");
            ioex.printStackTrace();
        }
    }

    private static void printInfo(String str) {
        System.out.println(str);
    }

    private void sendMessage(String str) {

        _out.println(str);
    }

    @Override
    public void run() {

        availableTowns.addAll(_towns);

        try (
                BufferedReader in = new BufferedReader(new InputStreamReader(_socket.getInputStream()));
                PrintWriter out = new PrintWriter(_socket.getOutputStream(), true)
            ){

            _in = in; _out = out;

            sendMessage("Let's play Goroda! You start, type "+ServerClientConstants.BYE+" to exit.");
            String line;

            while ((line = _in.readLine())!=null && !this.terminated) {
                this.terminated = line.trim().equals(ServerClientConstants.BYE);
                if(!this.terminated) {
                    printInfo("Received from " + _socket.getRemoteSocketAddress() + ": " + line);

                    if (playTown(line)) {
                        _score++;
                    }
                }
            }
            gameOver("Thank you for playing! Come again :)");
            printInfo("Client "+_socket.getRemoteSocketAddress()+" disconnected.");
        } catch (IOException ioex) {
            printInfo("Server failed in game loop\n\n");
        }

    }

    private boolean playTown(String town) {
        boolean result = town!=null;
        String serverTown;

        if (result){
            result = availableTowns.contains(town);
            if(!result){
                sendMessage("Your town "+town+" does not exist :( Try again!");
            } else {
                result = testTownBeginningMatchesEnd(town);
                if (!result) {
                    sendMessage("Your town must start with " + Character.toUpperCase(_lastChar) + ". You submitted: " + town);
                }
                if (result) {
                    if(submitTown(town)){
                        serverTown = makeServerTurn( "I won! no towns for you to enumerate!");
                        sendMessage("Accepted. AI turn: " + serverTown + ". Your town must start with: " + Character.toUpperCase(_lastChar));
                    } else {
                        gameOver("You won! All known to me towns has been enumerated!");
                    }
                }
            }
        }
        return result;
    }

    private void gameOver(String message) {
        sendMessage(message+" Your final score: "+_score);
        terminate();
    }

    private void terminate() {
        this.terminated = true;
    }

    private static boolean loadSet(Set towns, String citiesFile) {
        try{
            Files.lines(Paths.get(citiesFile), StandardCharsets.UTF_8).forEach(towns::add);
        } catch (IOException ioex) {
            printInfo("Couldn't read file: "+citiesFile+"\n\n");
            ioex.printStackTrace();
            System.exit(2);
        }
        return true;
    }
    private String makeServerTurn(String gameOverMessage){
        String serverTown = ((Optional<String>) availableTowns.stream().filter(item -> item.toString().length()>1 &&
                testTownBeginningMatchesEnd(item.toString())).findAny()).orElse(ServerClientConstants.EMPTY);
        if (!submitTown(serverTown)||serverTown.equals(ServerClientConstants.EMPTY))
            gameOver(gameOverMessage);
        testTownBeginningMatchesEnd(serverTown);//Just to correct _lastChar value after the server turn, ignore the return value
        return serverTown;
    }
    private boolean testTownBeginningMatchesEnd(String town){
        int shift = 1;
        if (_lastTown != null && town != null) {
            boolean excludedChar;
            do {
                _lastChar = _lastTown.charAt(_lastTown.length() - shift);
                if (shift == town.length()) break;
                excludedChar = ServerClientConstants.EXCLUDED.contains(Character.toLowerCase(_lastChar));
                if (excludedChar) shift++;
            } while (excludedChar);
            //printInfo(_lastTown+" "+_lastChar);
            return Character.toLowerCase(_lastChar) == Character.toLowerCase(town.charAt(0));
        } else return town != null;
    }
    private boolean submitTown(String town){
        availableTowns.remove(town);
        playedTowns.add(town);
        if (_lastTown != null) {
            _lastTown.ensureCapacity(town.length());
            _lastTown.replace(0, town.length(), town);
            _lastTown.setLength(town.length());
        } else {
            _lastTown = new StringBuffer(town);
        }
        return !availableTowns.isEmpty();
    }
}